/*
irc.c - FireTalk IRC protocol definitions
Copyright (C) 2000 Ian Gulliver

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "lclint.h"
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <strings.h>
#include <sys/time.h>
#include <time.h>

#include "firetalk.h"
#include "firetalk-int.h"
#include "irc.h"
#include "safestring.h"

#define IRC_SERVER "irc.openprojects.net"
#define IRC_PORT 6667
#define ROOMSTARTS "#+&"

typedef struct s_irc_whois {
	struct s_irc_whois *next;
	char *nickname;
	char *info;
	int flags;
	long idle;
} s_irc_whois;

typedef struct s_irc_connection {
	int s;
	char *nickname;
	char *password;
	char buffer[513];
	time_t lasttime;                                     /* last time we set idles */
	int isons;                                           /* number of ISON's we're waiting on replies to */
	struct s_irc_whois *whois_head;
	int passchange;                                      /* whether we are currently changing our pass */
	int usesilence;                                      /* are we on a network that understands SILENCE */
} s_irc_connection;

static char *irc_html_to_irc(const char * const string);
static char *irc_irc_to_html(const char * const string);
static int irc_send_printf(struct s_irc_connection * const c, const char * const format, ...);
static int irc_internal_disconnect(struct s_irc_connection * const c, const int error);
static char **irc_recv_parse(struct s_irc_connection * const c);

/*

from buddy structure:

tempint1 = ison, default 0, used during ISON checks
tempint2 = away, default 0, used during WHOIS checks

*/

static char *irc_html_to_irc(const char * const string) {
	static char *output = NULL;
	int o = 0;
	size_t l,i=0;
	l = strlen(string);
	/*@-onlytrans*/
	output = safe_realloc(output,l + 1);
	/*@+onlytrans*/
	/*@-infloops*/
	while (i < l) {
	/*@+infloops*/
		switch(string[i]) {
			case '&':
				/*@-unrecog*/
				if (!strncasecmp(&string[i],"&amp;",5)) {
					output[o++] = '&';
					i += 5;
				} else if (!strncasecmp(&string[i],"&gt;",4)) {
					output[o++] = '>';
					i += 4;
				} else if (!strncasecmp(&string[i],"&lt;",4)) {
					output[o++] = '<';
					i += 4;
				} else
					output[o++] = string[i++];
				break;
			case '<':
				if (!strncasecmp(&string[i],"<b>",3)) {
					output[o++] = (char) 2;
					i += 3;
				} else if (!strncasecmp(&string[i],"</b>",4)) {
					output[o++] = (char) 2;
					i += 4;
				} else if (!strncasecmp(&string[i],"<i>",3)) {
					output[o++] = (char) 22;
					i += 3;
				} else if (!strncasecmp(&string[i],"</i>",4)) {
					output[o++] = (char) 22;
					i += 4;
				} else if (!strncasecmp(&string[i],"<u>",3)) {
					output[o++] = (char) 31;
					i += 4;
				} else if (!strncasecmp(&string[i],"</u>",4)) {
					output[o++] = (char) 31;
					i += 5;
				} else
					output[o++] = string[i++];
				break;
				/*@+unrecog*/
			default:
				output[o++] = string[i++];
				break;
		}
	}
	output[o] = '\0';
	return output;
}

static char *irc_irc_to_html(const char * const string) {
	static char *output = NULL;
	int o = 0;
	size_t l,i=0;
	int inbold = 0, initalics = 0, inunderline = 0;
	l = strlen(string);
	/*@-onlytrans*/
	output = safe_realloc(output,(l * 5) + 1);
	/*@+onlytrans*/
	while (i < l) {
		switch(string[i]) {
			case 2:
				if (inbold == 1) {
					memcpy(&output[o],"</b>",4);
					o += 4;
					inbold = 0;
				} else {
					memcpy(&output[o],"<b>",3);
					o += 3;
					inbold = 1;
				}
				break;
			case 22:
				if (initalics == 1) {
					memcpy(&output[o],"</i>",4);
					o += 4;
					initalics = 0;
				} else {
					memcpy(&output[o],"<i>",3);
					o += 3;
					initalics = 1;
				}
				break;
			case 31:
				if (inunderline == 1) {
					memcpy(&output[o],"</u>",4);
					o += 4;
					inunderline = 0;
				} else {
					memcpy(&output[o],"<u>",3);
					o += 3;
					inunderline = 1;
				}
				break;
			case '&':
				memcpy(&output[o],"&amp;",5);
				o += 5;
				break;
			case '<':
				memcpy(&output[o],"&lt;",4);
				o += 4;
				break;
			case '>':
				memcpy(&output[o],"&gt;",4);
				o += 4;
				break;
			default:
				output[o++] = string[i];
				break;
		}
		i++;
	}
	output[o] = '\0';
	return output;
}

static int irc_internal_disconnect(struct s_irc_connection * const c, const int error) {
	struct s_irc_whois *whois_iter,*whois_iter2;
	/*@-unrecog*/ /* lclint can't parse my socket.h */
	shutdown(c->s, SHUT_RDWR);
	/*@+unrecog*/
	if (c->nickname)
		free(c->nickname);
	if (c->password)
		free(c->password);
	whois_iter = c->whois_head;
	while (whois_iter) {
		whois_iter2 = whois_iter->next;
		if (whois_iter->nickname != NULL)
			free(whois_iter->nickname);
		if (whois_iter->info != NULL)
			free(whois_iter->info);
		/*@-compdestroy*/
		free(whois_iter);
		/*@+compdestroy*/
		whois_iter = whois_iter2;
	}
	free(c);
	/*@-usereleased*/ /* firetalk_callback_disconnect only uses pointer, never derefs it */
	firetalk_callback_disconnect(c,error);
	/*@+usereleased*/
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

static int irc_send_printf(struct s_irc_connection * const c, const char * const format, ...) {
	va_list ap;
	size_t len,i,datai = 0;
	char data[513];
	char *tempchr;

	va_start(ap,format);

	len = strlen(format);
	for (i = 0; i < len; i++) {
		if (format[i] == '%') {
			switch (format[++i]) {
				case 's':
					tempchr = irc_html_to_irc(va_arg(ap, char *));
					if (datai + strlen(tempchr) > 509) {
						firetalkerror = FE_PACKETSIZE;
						return FE_PACKETSIZE;
					}
					safe_strncpy(&data[datai],tempchr,513-datai);
					datai += strlen(tempchr);
					break;
				case '%':
					data[datai++] = '%';
					break;
			}
		} else {
			data[datai++] = format[i];
			if (datai > 509) {
				firetalkerror = FE_PACKETSIZE;
				return FE_PACKETSIZE;
			}
		}
	}
	data[datai] = '\0';

	safe_strncat(data,"\r\n",513);
	datai = strlen(data);

	/*@-unrecog*/ /* lclint can't parse my socket.h */
	if (send(c->s,data,datai,0) != datai) {
	/*@+unrecog*/
		(void) irc_internal_disconnect(c,FE_PACKET);
		firetalkerror = FE_PACKET;
		return FE_PACKET;
	}
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

static char **irc_recv_parse(struct s_irc_connection * const c) {
	static char *args[256];
	size_t curarg;
	char *tempchr;
	char *tempchr2;
	static char data[513];
	char mydata[513];
	short length;

	curarg = strlen(c->buffer);

	args[0] = NULL;

	if (!strstr(c->buffer,"\r\n")) {
		/*@-unrecog*/ /* lclint can't parse my socket.h */
		if ((length = recv(c->s,&c->buffer[curarg],512 - curarg,0)) < 1) {
		/*@+unrecog*/
			(void) irc_internal_disconnect(c,FE_DISCONNECT);
			firetalkerror = FE_DISCONNECT;
			return args;
		}
		c->buffer[curarg + length] = '\0';
	}
	if ((tempchr = strstr(c->buffer,"\r\n"))) {
		tempchr[0] = '\0';
		if (c->buffer[0] == ':')
			safe_strncpy(data,&c->buffer[1],513);
		else
			safe_strncpy(data,c->buffer,513);
		safe_strncpy(mydata,&tempchr[2],513);
		safe_strncpy(c->buffer,mydata,513);
	} else {
		if (strlen(c->buffer) >= 512) {
			(void ) irc_internal_disconnect(c,FE_PACKETSIZE);
			firetalkerror = FE_PACKETSIZE;
			return NULL;
		}
		firetalkerror = FE_SUCCESS;
		return args;
	}

	curarg = 0;
	tempchr = data;

	while ((curarg < 256) && ((tempchr2 = strchr(tempchr,' ')) != NULL)) {
		args[curarg++] = tempchr;
		tempchr2[0] = '\0';
		tempchr = tempchr2 + 1;
		if (tempchr[0] == ':') {
			tempchr = irc_irc_to_html(&tempchr[1]);
			break;
		}
	}
	args[curarg++] = tempchr;
	args[curarg] = NULL;
	return args;
}

static char *irc_get_nickname(const char * const hostmask) {
	static char data[512];
	char *tempchr;
	safe_strncpy(data,hostmask,512);
	tempchr = strchr(data,'!');
	if (tempchr)
		tempchr[0] = '\0';
	return data;
}

int irc_set_nickname(void * const c, const char * const nickname) {
	return irc_send_printf(c,"NICK %s",nickname);
}

int irc_set_password(void * const c, const char * const newpass) {
	((struct s_irc_connection *)c)->passchange++;
	return irc_send_printf(c,"PRIVMSG NickServ :SET PASSWORD %s",newpass);
}

int irc_disconnect(void * const c) {
	(void) irc_send_printf(c,"QUIT :User disconnected");
	return irc_internal_disconnect(c,FE_USERDISCONNECT);
}

void *irc_connect_server(const char * server, short port, const char * const nickname, const char * const password) {
	struct s_irc_connection *c;
	struct hostent *he;
	struct in_addr addr;
	struct sockaddr_in fulladdr;
	char ourserver[sizeof(IRC_SERVER)] = IRC_SERVER;
	int flag;
	char **args;

	/*@-branchstate*/
	if (server == NULL)
		server = ourserver;
	if (port == 0)
		port = IRC_PORT;
	/*@+branchstate*/
	
	c = safe_malloc(sizeof(struct s_irc_connection));
	c->nickname = NULL;
	c->lasttime = 0;
	c->isons = 0;
	c->whois_head = NULL;
	c->password = safe_strdup(password);
	c->passchange = 0;
	c->usesilence = 1;

	/*@-unrecog*/ /* lclint can't parse my socket.h */
	if ( (c->s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
	/*@+unrecog*/
		(void) irc_internal_disconnect(c,FE_SOCKET);
		firetalkerror = FE_SOCKET;
		return NULL;
	}

	if (inet_aton(server,&addr) == 0) {
		/*@-unrecog*/ /* lclint can't parse my socket.h */
		he = gethostbyname(server);
		/*@+unrecog*/
		/*@-type*/
		if ((he == NULL) || (he->h_addr_list == NULL)) {
		/*@+type*/
			(void) irc_internal_disconnect(c,FE_RESOLV);
			firetalkerror = FE_RESOLV;
			return NULL;
		}
		/*@-type -nullderef*/
		memcpy(&addr,he->h_addr_list[0],sizeof(struct in_addr));
		/*@+type -nullderef*/
	}

	/*@-type -unrecog*/
	fulladdr.sin_family = AF_INET;
	/*@+type +unrecog*/
	/*@-type*/
	fulladdr.sin_port = htons(port);
	/*@+type*/
	memcpy(&fulladdr.sin_addr,&addr,sizeof(struct in_addr));

	/*@-unrecog*/
	if (connect(c->s,(const struct sockaddr *)&fulladdr,sizeof(struct sockaddr_in))) {
	/*@+unrecog*/
		(void) irc_internal_disconnect(c,FE_CONNECT);
		firetalkerror = FE_CONNECT;
		return NULL;
	}

	if (irc_send_printf(c,"USER %s %s %s :%s",nickname,server,server,nickname) != 0)
		return NULL;

	if (irc_send_printf(c,"NICK %s",nickname) != 0)
		return NULL;

	flag = 1;
	while (flag == 1) {
		args = irc_recv_parse(c);
		if (!args[0])
			continue;
		/* zero argument items */
		if (strcmp(args[0],"ERROR") == 0) {
			(void) irc_send_printf(c,"QUIT :error");
			(void) irc_internal_disconnect(c,FE_CONNECT);
			firetalkerror = FE_CONNECT;
			return NULL;
		}

		if (!args[1])
			continue;
		/* one argument items */
		if (strcmp(args[1],"ERROR") == 0) {
			(void) irc_send_printf(c,"QUIT :error");
			(void) irc_internal_disconnect(c,FE_CONNECT);
			firetalkerror = FE_CONNECT;
			return NULL;
		}
		if (strcmp(args[0],"PING") == 0) {
			if (irc_send_printf(c,"PONG %s",args[1]) != 0) {
				(void) irc_send_printf(c,"QUIT :error");
				(void) irc_internal_disconnect(c,FE_CONNECT);
				firetalkerror = FE_CONNECT;
				return NULL;
			}
		} else {
			switch (atoi(args[1])) {
				case 376:
					flag = 0;
					break;
				case 431:
				case 432:
				case 433:
				case 436:
				case 461:
					(void) irc_send_printf(c,"QUIT :Invalid nickname");
					(void) irc_internal_disconnect(c,FE_BADUSER);
					firetalkerror = FE_BADUSER;
					return NULL;
				case 465:
					(void) irc_send_printf(c,"QUIT :banned");
					(void) irc_internal_disconnect(c,FE_BLOCKED);
					firetalkerror = FE_BLOCKED;
					return NULL;
			}
		}
		
	}

	c->nickname = safe_strdup(nickname);

	return c;
}

void irc_signon(void * const c) {
	struct s_irc_connection *conn;
	conn = c;
	firetalk_callback_doinit(c,conn->nickname);
}

int irc_got_data(void * const c) {
	struct s_irc_connection *conn;
	char **args;
	char *tempchr, *tempchr2;
	struct s_firetalk_buddy *buddyiter;
	struct s_irc_whois *whoisiter, *whoisiter2;
	struct s_firetalk_handle *fchandle;
	int tempint,tempint2,tempint3;
	char tempbuf[512];

	fchandle = firetalk_find_handle(c);

	conn = c;

	args = irc_recv_parse(conn);
	while (args[0]) {
		/* zero argument items */
		if (strcmp(args[0],"PING") == 0) {
			if (args[1]) {
				if (irc_send_printf(c,"PONG %s",args[1]) != 0) {
					(void) irc_internal_disconnect(conn,FE_PACKET);
					firetalkerror = FE_PACKET;
					return FE_PACKET;
				}
			} else {
				if (irc_send_printf(c,"PONG") != 0) {
					(void) irc_internal_disconnect(conn,FE_PACKET);
					firetalkerror = FE_PACKET;
					return FE_PACKET;
				}
			}
		}
		if (args[1] != NULL) {
			tempint = atoi(args[1]);
			if (strcmp(args[1],"QUIT") == 0) {
				firetalk_callback_im_buddyonline(conn,irc_get_nickname(args[0]),0);
				if (irc_compare_nicks(conn->nickname,irc_get_nickname(args[0])) == 0)
					(void) irc_internal_disconnect(conn,FE_DISCONNECT);
				else
					firetalk_callback_chat_user_quit(conn,irc_get_nickname(args[0]));
			}
		}
		if ((args[1] != NULL) && (args[2] != NULL)) {
			/* two-arg commands */
			if (strcmp(args[1],"JOIN") == 0) {
				firetalk_callback_im_buddyonline(conn,irc_get_nickname(args[0]),1);
				if (irc_compare_nicks(conn->nickname,irc_get_nickname(args[0])) == 0)
					firetalk_callback_chat_joined(conn,args[2]);
				else
					firetalk_callback_chat_user_joined(conn,args[2],irc_get_nickname(args[0]));
			} else if (strcmp(args[1],"PART") == 0) {
				if (irc_compare_nicks(conn->nickname,irc_get_nickname(args[0])) == 0)
					firetalk_callback_chat_left(conn,args[2]);
				else
					firetalk_callback_chat_user_left(conn,args[2],irc_get_nickname(args[0]));
			} else if (strcmp(args[1],"NICK") == 0) {
				if (irc_compare_nicks(conn->nickname,irc_get_nickname(args[0])) == 0)
					firetalk_callback_newnick(conn,args[2]);
				else
					firetalk_callback_user_nickchanged(conn,irc_get_nickname(args[0]),args[2]);
			}
		}
		if ((args[1] != NULL) && (args[2] != NULL) && (args[3] != NULL)) {
			if (strcmp(args[1],"PRIVMSG") == 0) {
				/* scan for CTCP's */
				while ((tempchr = strchr(args[3],1))) {
					int correctctcp = 0;
					if ((tempchr2 = strchr(&tempchr[1],1))) {
						correctctcp = 1;
						tempchr2[0] = '\0';
					}
					/* we have a ctcp */
					if (strchr(ROOMSTARTS,args[2][0])) {
						/* chat room subcode */
						if (strncasecmp(&tempchr[1],"ACTION ",7) == 0)
							firetalk_callback_chat_getaction(conn,args[2],irc_get_nickname(args[0]),0,&tempchr[8]);
					} else {
						char *endcommand;
						endcommand = strchr(&tempchr[1],' ');
						if (endcommand) {
							*endcommand = '\0';
							endcommand++;
							firetalk_callback_subcode_request(conn,irc_get_nickname(args[0]),&tempchr[1],endcommand);
						} else
							firetalk_callback_subcode_request(conn,irc_get_nickname(args[0]),&tempchr[1],"");
					}
					if (correctctcp == 1)
						memcpy(tempchr,&tempchr2[1],strlen(&tempchr2[1]) + 1);
				}
				if (args[3][0] != '\0') {
					if (strchr(ROOMSTARTS,args[2][0]))
						firetalk_callback_chat_getmessage(conn,args[2],irc_get_nickname(args[0]),0,args[3]);
					else
						firetalk_callback_im_getmessage(conn,irc_get_nickname(args[0]),0,args[3]);
				}
			} else if (strcmp(args[1],"NOTICE") == 0) {
				/* scan for CTCP's */
				while ((tempchr = strchr(args[3],1))) {
					int correctctcp = 0;
					if ((tempchr2 = strchr(&tempchr[1],1))) {
						correctctcp = 1;
						tempchr2[0] = '\0';
					}
					/* we have a ctcp */
					if (strchr(ROOMSTARTS,args[2][0]) == NULL) {
						char *endcommand;
						endcommand = strchr(&tempchr[1],' ');
						if (endcommand) {
							*endcommand = '\0';
							endcommand++;
							firetalk_callback_subcode_reply(conn,irc_get_nickname(args[0]),&tempchr[1],endcommand);
						} else
							firetalk_callback_subcode_reply(conn,irc_get_nickname(args[0]),&tempchr[1],"");
					}
					if (correctctcp == 1)
						memcpy(tempchr,&tempchr2[1],strlen(&tempchr2[1]) + 1);
				}
				if (!strcasecmp(irc_get_nickname(args[0]),"NickServ")) {
					if ((conn->password != NULL) && (strstr(args[3],"IDENTIFY") != NULL))
						/* nickserv seems to be asking us to identify ourselves, and we have a password */
						if (irc_send_printf(conn,"PRIVMSG NickServ :IDENTIFY %s",conn->password) != 0) {
							(void) irc_internal_disconnect(conn,FE_PACKET);
							firetalkerror = FE_PACKET;
							return FE_PACKET;
						}
					if ((strstr(args[3],"Password changed") != NULL) && (conn->passchange != 0)) {
						/* successful change */
						conn->passchange--;
						firetalk_callback_passchanged(conn);
					}
					if ((strstr(args[3],"authentication required") != NULL) && (conn->passchange != 0)) {
						/* didn't log in with the right password initially, not happening */
						conn->passchange--;
						firetalk_callback_error(conn,FE_NOCHANGEPASS,NULL);
					}
					if ((strstr(args[3],"isn't registered") != NULL) && (conn->passchange != 0)) {
						/* nick not registered, fail */
						conn->passchange--;
						firetalk_callback_error(conn,FE_NOCHANGEPASS,NULL);
					}
				}
				if (args[3][0] != '\0') {
					if (strchr(ROOMSTARTS,args[2][0]))
						firetalk_callback_chat_getmessage(conn,args[2],irc_get_nickname(args[0]),1,args[3]);
					else
						firetalk_callback_im_getmessage(conn,irc_get_nickname(args[0]),1,args[3]);
				}
			} else if (strcmp(args[1],"TOPIC") == 0) {
				firetalk_callback_chat_gottopic(conn,args[2],args[3],irc_get_nickname(args[0]));
			} else if (strcmp(args[1],"KICK") == 0) {
				if (irc_compare_nicks(conn->nickname,irc_get_nickname(args[3])) == 0)
					firetalk_callback_chat_kicked(conn,args[2],irc_get_nickname(args[0]),args[4]);
				else {
					tempchr = safe_strdup(irc_get_nickname(args[0]));
					firetalk_callback_chat_user_kicked(conn,args[2],irc_get_nickname(args[3]),tempchr,args[4]);
					free(tempchr);
				}
			} else {
				switch (tempint) {
					case 301: /* RPL_AWAY */
						buddyiter = fchandle->buddy_head;
						while (buddyiter) {
							if (irc_compare_nicks(args[3],buddyiter->nickname) == 0) {
								buddyiter->tempint2 = 1;
								firetalk_callback_im_buddyaway(conn,args[3],1);
							}
							buddyiter = buddyiter->next;
						}
						break;
					case 303: /* RPL_ISON */
						tempchr = args[3];
						while ((tempchr != NULL) && (tempchr[0] != '\0')) {
							tempchr2 = strchr(tempchr,' ');
							if (tempchr2)
								tempchr2[0] = '\0';
							buddyiter = fchandle->buddy_head;
							while (buddyiter) {
								if (irc_compare_nicks(tempchr,buddyiter->nickname) == 0)
									buddyiter->tempint = 1;
								buddyiter = buddyiter->next;
							}
							if (tempchr2)
								tempchr = tempchr2 + 1;
							else
								tempchr = NULL;
						}
						if (--conn->isons <= 0) {
							conn->isons = 0;
							/* done, send the appropriate data */
							buddyiter = fchandle->buddy_head;
							while (buddyiter) {
								firetalk_callback_im_buddyonline(conn,buddyiter->nickname,buddyiter->tempint);
								if (buddyiter->tempint != 0) {
									buddyiter->tempint2 = 0; /* away */
									if (irc_send_printf(conn,"WHOIS %s",buddyiter->nickname) != 0) {
										(void) irc_internal_disconnect(conn,FE_PACKET);
										firetalkerror = FE_PACKET;
										return FE_PACKET;
									}
								}
								buddyiter = buddyiter->next;
							}
						}
						break;
					case 313: /* RPL_WHOISOPER */
						whoisiter = conn->whois_head;
						while (whoisiter) {
							if (irc_compare_nicks(args[3],whoisiter->nickname) == 0)
								whoisiter->flags |= FF_ADMIN;
							whoisiter = whoisiter->next;
						}
						break;
					case 318: /* RPL_ENDOFWHOIS */
						whoisiter = conn->whois_head;
						whoisiter2 = NULL;
						while (whoisiter) {
							if (irc_compare_nicks(args[3],whoisiter->nickname) == 0) {
								/* manual whois */
								firetalk_callback_gotinfo(conn,whoisiter->nickname,whoisiter->info,0,whoisiter->idle,whoisiter->flags);
								free(whoisiter->nickname);
								if (whoisiter->info)
									free(whoisiter->info);
								if (whoisiter2)
									whoisiter2->next = whoisiter->next;
								else
									conn->whois_head = whoisiter->next;
								free(whoisiter);
								break;
							}
							whoisiter2 = whoisiter;
							whoisiter = whoisiter->next;
						}
						buddyiter = fchandle->buddy_head;
						while (buddyiter) {
							if (irc_compare_nicks(args[3],buddyiter->nickname) == 0) {
								if (buddyiter->tempint2 == 0)
									firetalk_callback_im_buddyaway(conn,buddyiter->nickname,0);
							}
							buddyiter = buddyiter->next;
						}
						break;
					case 401: /* ERR_NOSUCHNICK */
					case 441: /* ERR_USERNOTINCHANNEL */
					case 443: /* ERR_USERONCHANNEL */
						firetalk_callback_im_buddyonline(conn,args[3],0);
						firetalk_callback_error(conn,FE_BADUSER,args[3]);
						if (!strcasecmp(args[3],"NickServ") && conn->passchange)
							conn->passchange--;
						whoisiter = conn->whois_head;
						whoisiter2 = NULL;
						while (whoisiter) {
							if (irc_compare_nicks(args[3],whoisiter->nickname) == 0) {
								free(whoisiter->nickname);
								if (whoisiter->info)
									free(whoisiter->info);
								/*@-branchstate*/
								if (whoisiter2)
									whoisiter2->next = whoisiter->next;
								else
									conn->whois_head = whoisiter->next;
								/*@+branchstate*/
								free(whoisiter);
								break;
							}
							whoisiter2 = whoisiter;
							whoisiter = whoisiter->next;
						}
						break;
					case 403: /* ERR_NOSUCHCHANNEL */
					case 442: /* ERR_NOTONCHANNEL */
						firetalk_callback_error(conn,FE_BADROOM,&args[3][1]);
						break;
					case 404: /* ERR_CANNOTSENDTOCHAN */
					case 405: /* ERR_TOOMANYCHANNELS */
					case 471: /* ERR_CHANNELISFULL */
					case 473: /* ERR_INVITEONLYCHAN */
					case 474: /* ERR_BANNEDFROMCHAN */
					case 475: /* ERR_BADCHANNELKEY */
						firetalk_callback_error(conn,FE_ROOMUNAVAILABLE,&args[3][1]);
						break;
					case 412: /* ERR_NOTEXTTOSEND */
						firetalk_callback_error(conn,FE_BADMESSAGE,NULL);
						break;
					case 421: /* ERR_UNKNOWNCOMMAND */
						if (strcmp(args[3],"SILENCE") == 0)
							conn->usesilence = 0;
						break;
					case 433: /* ERR_NICKNAMEINUSE */
						firetalk_callback_error(conn,FE_BADUSER,NULL);
						break;
					case 482: /* ERR_CHANOPRIVSNEEDED */
						firetalk_callback_error(conn,FE_NOPERMS,&args[3][1]);
						break;
				}
			}
		}
		if ((args[1] != NULL) && (args[2] != NULL) && (args[3] != NULL) && (args[4] != NULL)) {
			if (strcmp(args[1],"MODE") == 0) {
				tempint = 0;
				tempint2 = 4;
				tempint3 = 0;
				while ((args[tempint2] != NULL) && (args[3][tempint] != '\0')) {
					switch (args[3][tempint++]) {
						case '+':
							tempint3 = 1;
							break;
						case '-':
							tempint3 = -1;
							break;
						case 'o':
							if (tempint3 == 1)
								firetalk_callback_chat_user_opped(conn,args[2],args[tempint2++],irc_get_nickname(args[0]));
							else if (tempint3 == -1)
								firetalk_callback_chat_user_deopped(conn,args[2],args[tempint2++],irc_get_nickname(args[0]));
							break;
						default:
							tempint2++;
							break;
					}
				}
			} else {
				switch (tempint) {
					case 317: /* RPL_WHOISIDLE */
						whoisiter = conn->whois_head;
						while (whoisiter) {
							if (irc_compare_nicks(args[3],whoisiter->nickname) == 0)
								whoisiter->idle = atol(args[4])/60;
							whoisiter = whoisiter->next;
						}
						buddyiter = fchandle->buddy_head;
						while (buddyiter) {
							if (irc_compare_nicks(args[3],buddyiter->nickname) == 0)
								firetalk_callback_idleinfo(conn,args[3],atol(args[4]) / 60);
							buddyiter = buddyiter->next;
						}
						break;
					case 332: /* RPL_TOPIC */
						firetalk_callback_chat_gottopic(conn,args[3],args[4],NULL);
						break;
				}
			}
		}
		if ((args[1] != NULL) && (args[2] != NULL) && (args[3] != NULL) && (args[4] != NULL) && (args[5] != NULL)) {
			tempint = atoi(args[1]);
			if (tempint == 353) {
				tempchr2 = args[5];
				while ((tempchr = strchr(tempchr2,' '))) {
					tempchr[0] = '\0';
					if (tempchr2[0] == '@' || tempchr2[0] == '+') {
						firetalk_callback_chat_user_joined(conn,args[4],&tempchr2[1]);
						firetalk_callback_im_buddyonline(conn,&tempchr2[1],1);
						if (tempchr2[0] == '@')
							firetalk_callback_chat_user_opped(conn,args[4],&tempchr2[1],NULL);
					} else {
						firetalk_callback_chat_user_joined(conn,args[4],tempchr2);
						firetalk_callback_im_buddyonline(conn,tempchr2,1);
					}
					tempchr2 = tempchr + 1;
				}
			}
		}
		if ((args[1] != NULL) && (args[2] != NULL) && (args[3] != NULL) && (args[4] != NULL) && (args[5] != NULL) && (args[6] != NULL) && (args[7] != NULL)) {
			switch (tempint) {
				case 311: /* RPL_WHOISUSER */
					whoisiter = conn->whois_head;
					while (whoisiter) {
						if (irc_compare_nicks(args[3],whoisiter->nickname) == 0) {
							if (whoisiter->info)
								free(whoisiter->info);
							/*@-unrecog*/
							snprintf(tempbuf,512,"%s@%s: %s",args[4],args[5],args[7]);
							/*@+unrecog*/
							whoisiter->info = safe_strdup(tempbuf);
						}
						whoisiter = whoisiter->next;
					}
				break;
			}
		}
		
		if (strstr(conn->buffer,"\r\n"))
			args = irc_recv_parse(conn);
		else
			args[0] = NULL;
	}
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int irc_compare_nicks(const char * const nick1, const char * const nick2) {
	return strcasecmp(nick1,nick2);
}

int irc_get_fd(void * const c) {
	return ((struct s_irc_connection *)c)->s;
}

int irc_chat_join(void * const c, const char * const room) {
	return irc_send_printf(c,"JOIN %s",room);
}

int irc_chat_part(void * const c, const char * const room) {
	return irc_send_printf(c,"PART %s",room);
}

int irc_chat_send_message(void * const c, const char * const room, const char * const message, const int auto_flag) {
	if (auto_flag == 1)
		return irc_send_printf(c,"NOTICE %s :%s",room,message);
	else
		return irc_send_printf(c,"PRIVMSG %s :%s",room,message);
}

int irc_chat_send_action(void * const c, const char * const room, const char * const message, const int auto_flag) {
	if (auto_flag == 1)
		return irc_send_printf(c,"NOTICE %s :\001ACTION %s\001",room,message);
	else
		return irc_send_printf(c,"PRIVMSG %s :\001ACTION %s\001",room,message);
}

int irc_chat_invite(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"INVITE %s %s",who,room);
}

int irc_im_send_message(void * const c, const char * const dest, const char * const message, const int auto_flag) {
	if (auto_flag == 1)
		return irc_send_printf(c,"NOTICE %s :%s",dest,message);
	else
		return irc_send_printf(c,"PRIVMSG %s :%s",dest,message);
}

int irc_im_send_action(void * const c, const char * const dest, const char * const message, const int auto_flag) {
	if (auto_flag == 1)
		return irc_send_printf(c,"NOTICE %s :\001ACTION %s\001",dest,message);
	else
		return irc_send_printf(c,"PRIVMSG %s :\001ACTION %s\001",dest,message);
}

int irc_chat_set_topic(void * const c, const char * const room, const char * const topic) {
	return irc_send_printf(c,"TOPIC %s :%s",room,topic);
}

int irc_chat_op(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"MODE %s +o %s",room,who);
}

int irc_chat_deop(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"MODE %s -o %s",room,who);
}

int irc_chat_kick(void * const c, const char * const room, const char * const who, const char * const reason) {
	if (reason)
		return irc_send_printf(c,"KICK %s %s :%s",room,who,reason);
	else
		return irc_send_printf(c,"KICK %s %s",room,who);
}

int irc_im_add_buddy(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	conn = c;
	conn->isons++;
	return irc_send_printf(c,"ISON %s",nickname);
}

int irc_im_add_deny(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	conn = c;
	if (conn->usesilence == 1)
		return irc_send_printf(c,"SILENCE +%s!*@*",nickname);
	else
		return FE_SUCCESS;
}

int irc_im_remove_deny(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	conn = c;
	if (conn->usesilence == 1)
		return irc_send_printf(c,"SILENCE -%s!*@*",nickname);
	else
		return FE_SUCCESS;
}

int irc_get_info(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	struct s_irc_whois *whoistemp;
	conn = c;
	whoistemp = conn->whois_head;
	conn->whois_head = safe_malloc(sizeof(struct s_irc_whois));
	conn->whois_head->nickname = safe_strdup(nickname);
	conn->whois_head->flags = FF_NORMAL;
	conn->whois_head->idle = 0;
	conn->whois_head->info = NULL;
	conn->whois_head->next = whoistemp;
	return irc_send_printf(conn,"WHOIS %s",nickname);
}

int irc_set_away(void * const c, const char * const message) {
	if (message)
		return irc_send_printf(c,"AWAY :%s",message);
	else
		return irc_send_printf(c,"AWAY");
}

int irc_check_buddies(struct s_firetalk_handle * const c) {
	struct s_irc_connection *conn;
	struct s_firetalk_buddy *buddyiter;
	char obuf[1024];

	obuf[0] = '\0';

	conn = c->handle;
	if (conn->lasttime > (time(NULL) - 20))
		return FE_IDLEFAST;

	if (conn->isons > 0)
		return FE_IDLEFAST;

	buddyiter = c->buddy_head;
	while (buddyiter) {
		if (strlen(obuf) + strlen(buddyiter->nickname) > 502) {
			obuf[strlen(obuf)-1] = '\0';
			if (irc_send_printf(conn,"ISON %s",obuf) != 0) {
				(void) irc_internal_disconnect(conn,FE_PACKET);
				firetalkerror = FE_PACKET;
				return FE_PACKET;
			}
			conn->isons++;
			obuf[0] = '\0';
		}
		safe_strncat(obuf,buddyiter->nickname,1024);
		safe_strncat(obuf," ",1024);
		buddyiter->tempint = 0;
		buddyiter = buddyiter->next;
	}

	obuf[strlen(obuf)-1] = '\0';
	if (strlen(obuf) != 0) {
		if (irc_send_printf(conn,"ISON %s",obuf) != 0) {
			(void) irc_internal_disconnect(conn,FE_PACKET);
			firetalkerror = FE_PACKET;
			return FE_PACKET;
		}
		conn->isons++;
	}

	conn = c->handle;

	(void) time(&conn->lasttime);
	return FE_SUCCESS;
}

int irc_subcode_send_request(void * const c, const char * const to, const char * const command, const char * const args) {
	struct s_irc_connection *conn;
	conn = c;
	if (args == NULL) {
		if (irc_send_printf(conn,"PRIVMSG %s :\001%s\001",to,command) != 0) {
			(void) irc_internal_disconnect(conn,FE_PACKET);
			firetalkerror = FE_PACKET;
			return FE_PACKET;
		}
	} else {
		if (irc_send_printf(conn,"PRIVMSG %s :\001%s %s\001",to,command,args) != 0) {
			(void) irc_internal_disconnect(conn,FE_PACKET);
			firetalkerror = FE_PACKET;
			return FE_PACKET;
		}
	}
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int irc_subcode_send_reply(void * const c, const char * const to, const char * const command, const char * const args) {
	struct s_irc_connection *conn;
	conn = c;
	if (args == NULL) {
		if (irc_send_printf(conn,"NOTICE %s :\001%s\001",to,command) != 0) {
			(void) irc_internal_disconnect(conn,FE_PACKET);
			firetalkerror = FE_PACKET;
			return FE_PACKET;
		}
	} else {
		if (irc_send_printf(conn,"NOTICE %s :\001%s %s\001",to,command,args) != 0) {
			(void) irc_internal_disconnect(conn,FE_PACKET);
			firetalkerror = FE_PACKET;
			return FE_PACKET;
		}
	}
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

const char * const irc_normalize_room_name(const char * const name) {
	static char newname[2048];
	if (strchr(ROOMSTARTS,name[0]))
		return name;
	safe_strncpy(newname,"#",2048);
	safe_strncat(newname,name,2048);
	return newname;
}
